from loguru import logger
from canswim.model import CanswimModel
import gradio as gr
import duckdb


class ScanTab:

    def __init__(self, canswim_model: CanswimModel = None, db_path=None):
        assert canswim_model is not None
        assert db_path is not None
        self.canswim_model = canswim_model
        self.db_path = db_path
        with gr.Row():
            self.lowq = gr.Radio(
                choices=[80, 95, 99],
                value=80,
                label="Confidence level for lowest close price",
                info="Choose low price confidence percentage",
            )
            self.reward = gr.Radio(
                choices=[10, 15, 20, 25],
                value=20,
                label="Minimum probabilistic price gain.",
                info="Choose reward percentage",
            )
            self.rr = gr.Radio(
                choices=[3, 5, 10],
                value=3,
                label="Probabilistic Reward to Risk ratio between price increase and price drop within the forecast period.",
                info="Choose R/R ratio",
            )

        with gr.Row():
            self.scanBtn = gr.Button(value="Scan", variant="primary")

        with gr.Row():
            self.scanResult = gr.Dataframe()

        self.scanBtn.click(
            fn=self.scan_forecasts,
            inputs=[self.lowq, self.reward, self.rr],
            outputs=[self.scanResult],
        )

    def scan_forecasts(self, lowq, reward, rr):
        lq = (100 - lowq) / 100
        low_quantile_col = f"close_quantile_{lq}"
        mean_col = "close_quantile_0.5"
        with duckdb.connect(self.db_path) as db_con:
            sql_result = db_con.sql(
                f"""--sql
                SELECT 
                    f.symbol, 
                    f.start_date as forecast_start_date, 
                    arg_max(c.close, c.date) as prior_close_price, 
                    min("{low_quantile_col}") as forecast_close_low, 
                    max("{mean_col}") as forecast_close_high,
                    100*(forecast_close_high / prior_close_price - 1) as reward_percent,
                    (forecast_close_high - prior_close_price)/GREATEST(prior_close_price-forecast_close_low, 0.01) as reward_risk,
                    max(e.mal_error) as backtest_error,
                    max(c.date) as prior_close_date,
                FROM forecast f, close_price c, backtest_error as e, latest_forecast as lf
                WHERE f.symbol = lf.symbol AND
                    f.symbol = c.symbol AND f.symbol = e.symbol AND c.date < lf.date
                GROUP BY f.symbol, f.start_date, c.symbol, e.symbol, lf.symbol, lf.date
                HAVING forecast_close_high > prior_close_price AND
                    f.start_date = lf.date AND
                    reward_risk> $rr AND reward_percent >= $reward
                    and forecast_start_date = max(lf.date)
                """,
                params={
                    "rr": rr,
                    "reward": reward,
                },
            )
            logger.info(f"SQL Result: \n{sql_result}")
            df = sql_result.df()
            df["prior_close_date"] = df["prior_close_date"].dt.strftime("%Y-%m-%d")
            df["forecast_start_date"] = df["forecast_start_date"].dt.strftime(
                "%Y-%m-%d"
            )
            logger.debug(f"df types: {df.dtypes}")
            # dateformat = lambda d: d.strftime("%d %b, %Y")
            df_styler = df.style.format(
                # {"prior_close_date": dateformat, "forecast_start_date": dateformat},
                precision=2,
                thousands=",",
                decimal=".",
            )
            # .format(...date columns format ... .format(na_rep='PASS', precision=2, subset=[1, 2]))
            return df_styler



"""
Forecast stock price movement and upload results to HF Hub
"""

import pandas as pd
from loguru import logger
from canswim.model import CanswimModel
from canswim.hfhub import HFHub
from pandas.tseries.offsets import BDay
from loguru import logger
import os
from canswim import constants
from typing import List
import duckdb


class CanswimForecaster:
    def __init__(self):
        self.data_dir = os.getenv("data_dir", "data")
        self.data_3rd_party = os.getenv("data-3rd-party", "data-3rd-party")
        self.stock_tickers_list = os.getenv("stock_tickers_list", "all_stocks.csv")
        logger.info(f"Stocks train list: {self.stock_tickers_list}")
        self.n_stocks = int(os.getenv("n_stocks", 50))
        logger.info(f"n_stocks: {self.n_stocks}")
        self.forecast_subdir = os.getenv("forecast_subdir", "forecast/")
        logger.info(f"Forecast data path: {self.forecast_subdir}")
        self.canswim_model = CanswimModel()
        self.hfhub = HFHub()

    def download_model(self):
        """Load model from HF Hub"""
        # download model from hf hub
        self.canswim_model.download_model()
        logger.info("trainer params", self.canswim_model.torch_model.trainer_params)
        self.canswim_model.torch_model.trainer_params["logger"] = False

    def load_model(self):
        """Load model from local storage"""
        self.canswim_model.load()
        logger.info("trainer params", self.canswim_model.torch_model.trainer_params)
        self.canswim_model.torch_model.trainer_params["logger"] = False

    def download_data(self):
        """Prepare time series for model forecast"""
        self.hfhub.download_data()
        # load raw data from hf hub
        self.start_date = pd.Timestamp.now() - BDay(
            # min_samples is not sufficient as it does not account for all market days without price data
            # such as holidays. Also for some sparse data sets such as analyst estimates, there are periods of
            # 2 or more years without data.
            # Add a sufficiently big sample pad to include all market off days.
            n=self.canswim_model.min_samples
            * 2
        )

    def get_forecast(self, forecast_start_date: pd.Timestamp = None):
        logger.info("Forecast start. Calling model predict().")
        forecasted_tickers = []
        target_sliced_list = []
        past_cov_list = []
        future_cov_list = []
        tickers_list = self.canswim_model.targets_ticker_list
        # trim end of targets to specified forecast start date
        if forecast_start_date is not None:
            logger.debug(
                f"Dropping target samples after forecast start date, included: {forecast_start_date}"
            )
            for i, ts in enumerate(self.canswim_model.targets_list):
                try:
                    logger.debug(
                        f"Target {tickers_list[i]} start date, end date, sample count: {ts.start_time()}, {ts.end_time()}, {len(ts)}"
                    )
                    # if forecast start date is after the end of the target time series,
                    # then we can still forecast if the target series ends on the business day before the forecast start date
                    cutoff_forecast_start_date = ts.end_time() + BDay(n=1)
                    if forecast_start_date == cutoff_forecast_start_date:
                        target_sliced = ts
                    if forecast_start_date < cutoff_forecast_start_date:
                        target_sliced = ts.drop_after(forecast_start_date)
                    # we can only provide forecasts
                    # when forecast start date is immediately after target series end date
                    # and there are sufficient number of historical data samples
                    if (
                        forecast_start_date <= cutoff_forecast_start_date
                        and len(ts) >= self.canswim_model.min_samples
                    ):
                        forecasted_tickers.append(tickers_list[i])
                        target_sliced_list.append(target_sliced)
                        past_cov_list.append(self.canswim_model.past_cov_list[i])
                        future_cov_list.append(self.canswim_model.future_cov_list[i])
                    else:
                        logger.info(
                            f"Skipping {tickers_list[i]} for forecast start date {forecast_start_date} due to lack of historical data."
                        )
                except ValueError as e:
                    logger.warning(
                        f"Skipping {tickers_list[i]} for forecast start date {forecast_start_date} due to error: {type(e)}: {e}"
                    )
        if len(target_sliced_list) > 0:
            canswim_forecast = self.canswim_model.predict(
                target=target_sliced_list,
                past_covariates=past_cov_list,
                future_covariates=future_cov_list,
            )
            logger.info("Forecast finished.")
            forecasts = dict(zip(forecasted_tickers, canswim_forecast))
            return forecasts
        else:
            logger.warning(
                "No stocks have enough data in this batch. Skipping forecast."
            )
            return None

    def _get_stocks_without_forecast(self, stocks_df=None, forecast_start_date=None):
        if forecast_start_date is not None:
            dt = pd.Timestamp(forecast_start_date)
        else:
            dt = pd.Timestamp.now()
        # align date to closest business day
        # leave as is if dt is a business day,
        # otherwise move forward to next business day
        bd = dt + 0 * BDay()
        y = bd.year
        m = bd.month
        d = bd.day
        # logger.debug(f"Forecast start date, year, month, day: {bd}, {y}, {m}, {d}")
        # logger.debug(f"len(stocks_df): {len(stocks_df)}")
        # logger.debug(f"stocks_df: {stocks_df}")
        df = duckdb.sql(
            f"""--sql
            CREATE OR REPLACE TABLE stock_group AS SELECT Symbol from stocks_df;
            SELECT symbol, count(*), forecast_start_year, forecast_start_month, forecast_start_day
            FROM read_parquet('{self.data_dir}/{self.forecast_subdir}**/*.parquet', hive_partitioning = 1) as f
            SEMI JOIN stock_group
            ON f.symbol = stock_group.symbol
            GROUP BY f.symbol, forecast_start_year, forecast_start_month, forecast_start_day
            HAVING 
                forecast_start_year={y} AND
                forecast_start_month={m} AND
                forecast_start_day={d} AND
                count(*) >= {self.canswim_model.pred_horizon}
            """
        ).df()
        # logger.debug(f"sql result: {df}")
        stocks_with_saved_forecast = set(df["symbol"])
        logger.debug(
            f"""These stocks already have a saved forecast: {stocks_with_saved_forecast}"""
        )
        stocks_without_forecast = set(stocks_df["Symbol"]) - stocks_with_saved_forecast
        stocks_without_forecast = sorted(list(stocks_without_forecast))
        logger.debug(
            f"""These stocks do not have a saved forecast yet: {stocks_without_forecast}"""
        )
        return stocks_without_forecast

    def prep_next_stock_group(self, forecast_start_date=None):
        """Generator which iterates over all stocks and prepares them in groups."""
        stocks_file = f"{self.data_dir}/{self.data_3rd_party}/{self.stock_tickers_list}"
        logger.info(f"Loading stock tickers from {stocks_file}.")
        all_stock_tickers = pd.read_csv(stocks_file)
        # trim any duplicate tickers
        all_stock_tickers.index = all_stock_tickers.index.drop_duplicates()
        # drop any empty symbols/tickers
        all_stock_tickers = all_stock_tickers[all_stock_tickers.index.notnull()]
        logger.info(f"Loaded {len(all_stock_tickers)} symbols in total")
        stock_list = self._get_stocks_without_forecast(
            stocks_df=all_stock_tickers, forecast_start_date=forecast_start_date
        )
        logger.info(
            f"{len(all_stock_tickers)-len(stock_list)} stock already have forecast saved."
        )
        logger.info(f"{len(stock_list)} stock tickers candidates for new forecast.")
        if self.n_stocks < 0 or self.n_stocks > len(stock_list):
            self.n_stocks = len(stock_list)
        # group tickers in workable sample sizes for each forecast pass
        # credit ref: https://stackoverflow.com/questions/434287/how-to-iterate-over-a-list-in-chunks
        for pos in range(0, len(stock_list), self.n_stocks):
            stock_group = stock_list[pos : pos + self.n_stocks]
            self.canswim_model.load_data(
                stock_tickers=stock_group, start_date=self.start_date
            )
            # prepare timeseries for forecast
            self.canswim_model.prepare_forecast_data(start_date=self.start_date)
            logger.info(f"Prepared forecast data for {len(stock_group)}: {stock_group}")
            yield pos

    def save_forecast(self, forecasts: dict = None):
        """Saves forecast data to local database"""

        def _list_to_df(forecasts: dict = None):
            """Format list of forecasts as a dataframe to be saved as a partitioned parquet dir"""
            forecast_df = pd.DataFrame()
            for t, ts in forecasts.items():
                pred_start = ts.start_time()
                # logger.debug(f"Next forecast timeseries: {ts}")
                # normalize name of target series column if needed (e.g. "Adj Close" -> "Close")
                if self.canswim_model.target_column != "Close":
                    ts = ts.with_columns_renamed(
                        self.canswim_model.target_column, "Close"
                    )
                # convert probabilistic forecast into a dataframe with quantile samples as columns Close_01, Close_02...
                df = ts.pd_dataframe()
                # save commonly used quantiles
                for q in constants.quantiles:
                    qseries = df.quantile(q=q, axis=1)
                    qname = f"close_quantile_{q}"
                    df[qname] = qseries
                df["symbol"] = t
                df["forecast_start_year"] = pred_start.year
                df["forecast_start_month"] = pred_start.month
                df["forecast_start_day"] = pred_start.day
                # logger.debug(f"Next forecast sample: {df}")
                forecast_df = pd.concat([forecast_df, df])
            return forecast_df

        assert forecasts is not None and len(forecasts) > 0
        forecast_df = _list_to_df(forecasts)
        logger.info(
            f"Saving forecast_df with {len(forecast_df.columns)} columns, {len(forecast_df)} rows: {forecast_df}"
        )
        forecast_df.to_parquet(
            f"{self.data_dir}/{self.forecast_subdir}",
            partition_cols=[
                "symbol",
                "forecast_start_year",
                "forecast_start_month",
                "forecast_start_day",
            ],
        )
        logger.info(f"Saved forecast data to: {self.forecast_subdir}")

    def upload_data(self):
        self.hfhub.upload_data()


# main function
def main(forecast_start_date: str = None):
    logger.info("Running forecast on stocks and uploading results to HF Hub...")
    if forecast_start_date is not None:
        logger.info(f"forecast_start_date: {forecast_start_date}")
        forecast_start_date = pd.Timestamp(forecast_start_date)
    cf = CanswimForecaster()
    cf.download_model()
    cf.download_data()
    ## loop in groups over all stocks
    # next(cf.prep_next_stock_group())
    for pos in cf.prep_next_stock_group(forecast_start_date=forecast_start_date):
        forecasts = cf.get_forecast(forecast_start_date=forecast_start_date)
        ## save new or update existing data file
        if forecasts:
            cf.save_forecast(forecasts)
    cf.upload_data()
    logger.info("Finished forecast task.")


if __name__ == "__main__":
    main()



from pandas.tseries.offsets import BDay
import pandas as pd
from darts import TimeSeries
from darts.dataprocessing.transformers import MissingValuesFiller
from typing import Union
import numpy as np
from loguru import logger
import os

fiscal_periods = ["quarter", "annual"]
fiscal_freq = {"annual": "Y", "quarter": "Q"}


# credit for implementation: https://stackoverflow.com/a/39068260/12015435
def is_business_day(date):
    return bool(len(pd.bdate_range(date, date)))


# align date to nearest business day
def to_biz_day(date=None, report_time=None):
    if not is_business_day(date):
        # if report time is After Market Close, move to nearest prior business day
        if report_time == 1 or report_time == "amc":
            return date - BDay(n=1)
        # otherwise move to nearest future business day
        else:
            return date + BDay(n=1)
    else:
        return date


class Covariates:
    def __init__(self):
        self.past_covariates = {}
        self.future_covariates = {}
        self.data_dir = os.getenv("data_dir", "data")
        self.data_3rd_party = os.getenv("data-3rd-party", "data-3rd-party")

    @property
    def pyarrow_filters(self):
        return [
            ("Symbol", "in", self.__load_tickers),
            ("Date", ">=", self.__start_date),
        ]

    def get_price_covariates(self, stock_price_series=None, target_columns=None):
        logger.info("preparing past covariates: price and volume")
        # drop columns used in target series
        past_price_covariates = {
            t: stock_price_series[t].drop_columns(col_names=target_columns)
            for t in stock_price_series.keys()
        }
        return past_price_covariates

    # backfill quarterly earnigs and revenue estimates so that the model can see the next quarter's estimates during the previous quarter days

    def back_fill_earn_estimates(self, t_earn=None):
        t_earn["time"] = t_earn["time"].bfill()
        t_earn["epsEstimated"] = t_earn["epsEstimated"].bfill()
        t_earn["revenueEstimated"] = t_earn["revenueEstimated"].bfill()
        t_earn["fiscalDateEnding_day"] = t_earn["fiscalDateEnding_day"].bfill()
        t_earn["fiscalDateEnding_month"] = t_earn["fiscalDateEnding_month"].bfill()
        t_earn["fiscalDateEnding_year"] = t_earn["fiscalDateEnding_year"].bfill()
        return t_earn

    def align_earn_to_business_days(self, t_earn=None):
        assert not t_earn.index.isnull().any()
        new_index = t_earn.index.map(
            lambda x: to_biz_day(date=x, report_time=t_earn.at[x, "time"])
        )
        t_earn.index = new_index
        if t_earn.index.isnull().any():
            logger.info(t_earn[t_earn.index.isnull()])
        for i in t_earn.index:
            assert is_business_day(i)
        return t_earn

    def prepare_earn_series(self, tickers=None):
        logger.info("preparing past covariates: earnings estimates ")
        # convert date strings to numerical representation
        earn_df = self.earnings_loaded_df.copy()
        # logger.info("self.earnings_loaded_df.columns", self.earnings_loaded_df.columns)
        ufd = pd.to_datetime(earn_df["updatedFromDate"])
        ufd_year = ufd.dt.year
        ufd_month = ufd.dt.month
        ufd_day = ufd.dt.day

        earn_n_cols = len(earn_df.columns)
        earn_df.insert(loc=earn_n_cols, column="updatedFromDate_year", value=ufd_year)
        earn_df.insert(loc=earn_n_cols, column="updatedFromDate_month", value=ufd_month)
        earn_df.insert(loc=earn_n_cols, column="updatedFromDate_day", value=ufd_day)
        earn_df.pop("updatedFromDate")

        # convert date strings to numerical representation
        fde = pd.to_datetime(earn_df["fiscalDateEnding"])
        fde_year = fde.dt.year
        fde_month = fde.dt.month
        fde_day = fde.dt.day

        earn_n_cols = len(earn_df.columns)
        earn_df.insert(loc=earn_n_cols, column="fiscalDateEnding_year", value=fde_year)
        earn_df.insert(
            loc=earn_n_cols, column="fiscalDateEnding_month", value=fde_month
        )
        earn_df.insert(loc=earn_n_cols, column="fiscalDateEnding_day", value=fde_day)
        earn_df.pop("fiscalDateEnding")

        # convert earnings reporting time - Before Market Open / After Market Close - categories to numerical representation
        earn_df["time"] = pd.Categorical(
            earn_df["time"], categories=["bmo", "amc"]
        ).codes
        # earn_df["time"] = (
        #     earn_df["time"]
        #     .replace(["bmo", "amc", "--", "dmh"], [0, 1, -1, -1], inplace=False)
        #     .astype("int32")
        # )
        # convert earnings dataframe to series
        t_earn_series = {}
        for t in list(tickers):
            try:
                # logger.info(f'ticker: {t}')
                t_earn = earn_df.loc[[t]].copy()
                t_earn = t_earn.droplevel("Symbol")
                t_earn.index = pd.to_datetime(t_earn.index)
                # logger.info(f'index type for {t}: {type(t_earn.index)}')
                assert not t_earn.index.duplicated().any()
                assert not t_earn.index.isnull().any()
                t_earn = self.align_earn_to_business_days(t_earn)
                # drop rows with duplicate datetime index values
                t_earn = t_earn[~t_earn.index.duplicated(keep="last")]
                # t_earn = (
                #     t_earn.reset_index()
                #     .drop_duplicates(subset="Date", keep="last")
                #     .set_index("Date")
                # )
                # logger.info(f't_earn freq: {t_earn.index}')
                tes_tmp = TimeSeries.from_dataframe(
                    t_earn, freq="B", fill_missing_dates=True
                )
                t_earn = self.back_fill_earn_estimates(t_earn=tes_tmp.pd_dataframe())
                tes = TimeSeries.from_dataframe(t_earn, fillna_value=-1)
                assert len(tes.gaps()) == 0
                t_earn_series[t] = tes
            except KeyError as e:
                logger.warning(f"Skipping {t} due to error: {type(e)}: {e}")

        return t_earn_series

    def prepare_institutional_symbol_ownership_series(self, stock_price_series=None):
        logger.info(f"preparing past covariates: institutional ownership of symbol")
        inst_ownership_df = self.inst_symbol_ownership_df.copy()
        # cleanup data with known dirty columns
        inst_ownership_df["cik"] = (
            pd.to_numeric(inst_ownership_df["cik"], errors="coerce")
            .fillna(0)
            .astype(pd.Int64Dtype())
        )
        inst_ownership_df["totalCallsChange"] = (
            pd.to_numeric(inst_ownership_df["totalCallsChange"], errors="coerce")
            .fillna(0)
            .astype(pd.Float64Dtype())
        )
        inst_ownership_df["totalPutsChange"] = (
            pd.to_numeric(inst_ownership_df["totalPutsChange"], errors="coerce")
            .fillna(0)
            .astype(pd.Float64Dtype())
        )
        # convert earnings dataframe to series
        t_inst_ownership_series = {}
        for t, prices in stock_price_series.items():
            try:
                # logger.info(f"ticker: {t}")
                t_iown = inst_ownership_df.loc[[t]]
                t_iown = t_iown.droplevel("Symbol")
                t_iown.index = pd.to_datetime(t_iown.index)
                # logger.info(f"t_iown index: {t_iown.index}")
                # logger.info(f"t_iown index to biz days")
                # assert not t_iown.index.duplicated().any(), "date index has duplicates"
                # logger.info(f"t_iown no index duplicates")
                # assert not t_iown.index.isnull().any(), "date index has missing values"
                # logger.info(f"t_iown no index NaNs")
                t_iown = self.df_index_to_biz_days(t_iown)
                # logger.info(f"t_iown index to biz days")
                # t_iown = (
                #     t_iown.reset_index()
                #     .drop_duplicates(subset="Date", keep="last")
                #     .set_index("Date")
                # )
                t_iown = t_iown[~t_iown.index.duplicated(keep="last")]
                assert t_iown.index.is_unique, "date index has duplicates"
                # logger.info(f"t_iown no index duplicates")
                assert not t_iown.index.isnull().any(), "date index has missing values"
                # logger.info(f"t_iown no index NaNs")
                # logger.info(f't_earn freq: {t_earn.index}')
                # save cik as a static covariate
                ##cik = t_iown["cik"].iloc[0].astype(pd.Int64Dtype)
                ##t_iown = t_iown.drop(columns=["cik"])
                ##static_covs_single = pd.DataFrame(data={"cik": [cik]})
                # logger.info(f"Company with ticker {t} has cik: {cik}")
                ts_tmp = TimeSeries.from_dataframe(
                    t_iown, freq="B", fill_missing_dates=True
                )
                t_iown = ts_tmp.pd_dataframe()
                t_iown.ffill(inplace=True)
                # use 0 as a filler for unknown ownership absolute values and deltas
                ts = TimeSeries.from_dataframe(t_iown, fillna_value=0)
                ts_padded = self.pad_covs(
                    cov_series=ts, price_series=prices, fillna_value=0
                )
                assert (
                    len(ts_padded.gaps()) == 0
                ), f"found gaps in series: \n{ts_padded.gaps()}"
                t_inst_ownership_series[t] = ts_padded
            except KeyError as e:
                logger.info(
                    f"""Skipping {t} from covariates series. 
                        No institutional ownership data available for {t}, error: {type(e)}, {e}"""
                )
            except AssertionError as e:
                logger.warning(f"Skipping {t} due to error: {type(e)}: {e}")

        return t_inst_ownership_series

    def stack_covariates(self, old_covs=None, new_covs=None, min_samples=1):
        logger.info(f"stacking covariates.")
        # stack sales and earnigns to past covariates
        stacked_covs = {}
        for t, covs in list(old_covs.items()):
            try:
                assert (
                    type(new_covs[t]) == TimeSeries
                ), f"type of {t} is not TimeSeries, but {type(new_covs[t])}"
                # logger.info(f'stacking future covs for {t}')
                old_sliced = covs.slice_intersect(new_covs[t])
                new_sliced = new_covs[t].slice_intersect(old_sliced)
                stacked = old_sliced.stack(new_sliced)
                # logger.debug(
                #     f"covariates for {t} start time: {stacked.start_time()}, end time: {stacked.end_time()}"
                # )
                # logger.debug(f"covariates for {t} sample: \n{stacked.pd_dataframe()}")
                if len(stacked) >= min_samples:
                    stacked_covs[t] = stacked
                else:
                    logger.info(
                        f"Skipping {t} due to lack of min {min_samples} samples. Only {len(stacked)} samples available in covs stack."
                    )
            except KeyError as e:
                logger.warning(f"Skipping {t} due to error: {type(e)}: {e}")
        if len(stacked_covs.keys()) > 0:
            logger.info(f"stacked covariates column count: {len(stacked.columns)}")
        return stacked_covs

    def df_index_to_biz_days(self, df=None):
        new_index = df.index.map(lambda x: to_biz_day(date=x))
        df.index = new_index
        return df

    def pad_covs(self, cov_series=None, price_series=None, fillna_value=-1):
        """
        Pad a ticker's covariate series to align with target price series
        """
        updated_cov_series = None
        if cov_series.end_time() < price_series.end_time():
            df = cov_series.pd_dataframe()
            new_cov_df = df.reindex(
                price_series.pd_dataframe().index, method="ffill", copy=True
            )
            new_cov_ser = TimeSeries.from_dataframe(
                new_cov_df, freq="B", fillna_value=fillna_value
            )
            updated_cov_series = new_cov_ser
        else:
            updated_cov_series = cov_series
        return updated_cov_series

    def prepare_key_metrics(self, stock_price_series=None):
        logger.info("preparing past covariates: key metrics")
        kms_loaded_df = self.kms_loaded_df.copy()
        # logger.info(kms_loaded_df)
        kms_loaded_df = kms_loaded_df[~kms_loaded_df.index.duplicated(keep="first")]
        assert kms_loaded_df.index.is_unique
        kms_loaded_df.index
        len(kms_loaded_df.index)
        # kms_loaded_df["date"] = pd.to_datetime(kms_loaded_df["date"])
        kms_unique = kms_loaded_df.drop_duplicates()  # subset=["symbol", "date"])
        assert not kms_unique.duplicated().any()
        # kms_unique = kms_unique.set_index(keys=["symbol", "date"])
        assert not kms_unique.index.has_duplicates
        kms_loaded_df = kms_unique.copy()
        # convert earnings reporting time - Before Market Open / After Market Close - categories to numerical representation
        kms_loaded_df["period"] = pd.Categorical(
            kms_loaded_df["period"], categories=["_", "Q1", "Q2", "Q3", "Q4"]
        ).codes
        # kms_loaded_df["period"] = (
        #     kms_loaded_df["period"]
        #     .replace(["Q1", "Q2", "Q3", "Q4"], [1, 2, 3, 4], inplace=False)
        #     .astype("int32")
        # )

        t_kms_series = {}
        for t, prices in stock_price_series.items():
            try:
                # logger.info(f"ticker {t}")
                kms_df = kms_loaded_df.loc[[t]].copy()
                # logger.info(f'ticker_series[{t}] start time, end time: {ticker_series[t].start_time()}, {ticker_series[t].end_time()}')
                # logger.info(f'kms_ser_df start time, end time: {kms_df.index[0]}, {kms_df.index[-1]}')
                kms_df = kms_df.droplevel("Symbol")
                kms_df.index = pd.to_datetime(kms_df.index)
                # logger.info(f'index type for {t}: {type(t_kms.index)}')
                assert kms_df.index.is_unique
                kms_df = kms_df.dropna()
                # logger.info("kms_df\n", kms_df[kms_df.isnull()])
                assert not kms_df.isnull().values.any()
                assert len(kms_df) > 0, f"No key metrics available for {t}"
                # logger.info(f'{t} earnings: \n{t_kms.columns}')
                kms_df = self.df_index_to_biz_days(kms_df)
                tkms_series_tmp = TimeSeries.from_dataframe(
                    kms_df, freq="B", fill_missing_dates=True
                )
                # logger.info(f'kms_series_tmp start time, end time: {tkms_series_tmp.start_time()}, {tkms_series_tmp.end_time()}')
                kms_df_ext = tkms_series_tmp.pd_dataframe()
                kms_df_ext.ffill(inplace=True)
                kms_ser = TimeSeries.from_dataframe(kms_df, freq="B", fillna_value=-1)
                kms_ser_padded = self.pad_covs(cov_series=kms_ser, price_series=prices)
                # logger.info(f'kms_ser_padded start time, end time: {kms_ser_padded.start_time()}, {kms_ser_padded.end_time()}')
                assert (
                    len(kms_ser_padded.gaps()) == 0
                ), f"found gaps in tmks series: \n{kms_ser_padded.gaps()}"
                t_kms_series[t] = kms_ser_padded
            except (KeyError, AssertionError) as e:
                logger.warning(f"Skipping {t} due to error: {type(e)}: {e}")
        # logger.info("t_kms_series:", t_kms_series)
        return t_kms_series

    def prepare_broad_market_series(self, train_date_start=None):
        logger.info("preparing past covariates: broad market indexes")
        broad_market_df = self.broad_market_df.copy()
        # flatten column hierarchy so Darts can use as covariate series
        broad_market_df.columns = [f"{i}_{j}" for i, j in broad_market_df.columns]
        # CBOE VIX volatility, DYX USD and TNT 10Y Treasury indices do not have meaningful values for Volume
        broad_market_df = broad_market_df.drop(
            columns=["^VIX_Volume", "DX-Y.NYB_Volume", "^TNX_Volume"]
        )
        # fix datetime index type issue
        # https://stackoverflow.com/questions/48248239/pandas-how-to-convert-rangeindex-into-datetimeindex
        broad_market_df.index = pd.to_datetime(broad_market_df.index)
        broad_market_series = TimeSeries.from_dataframe(broad_market_df, freq="B")
        broad_market_series = broad_market_series.slice(
            train_date_start, broad_market_series.end_time()
        )
        filler = MissingValuesFiller(n_jobs=-1)
        series_filled = filler.transform(broad_market_series)
        assert len(series_filled.gaps()) == 0
        broad_market_series = series_filled
        return broad_market_series

    def prepare_sectors_series(self, train_date_start=None):
        logger.info("preparing past covariates: market sectors")
        sectors_df = self.sectors_df.copy()
        # flatten column hierarchy so Darts can use as covariate series
        sectors_df.columns = [f"{i}_{j}" for i, j in sectors_df.columns]
        # fix datetime index type issue
        # https://stackoverflow.com/questions/48248239/pandas-how-to-convert-rangeindex-into-datetimeindex
        sectors_df.index = pd.to_datetime(sectors_df.index)
        sectors_series = TimeSeries.from_dataframe(sectors_df, freq="B")
        sectors_series = sectors_series.slice(
            train_date_start, sectors_series.end_time()
        )
        filler = MissingValuesFiller(n_jobs=-1)
        series_filled = filler.transform(sectors_series)
        assert len(series_filled.gaps()) == 0
        sectors_series = series_filled
        # logger.debug(
        #     f"Finished preparing past covariates: market sectors. {len(sectors_series)} records, columns: {sectors_series.columns}"
        # )
        return sectors_series

    def prepare_industry_fund_series(self, train_date_start=None):
        logger.info("preparing past covariates: industry funds")
        industry_funds_df = self.industry_funds_df.copy()
        # flatten column hierarchy so Darts can use as covariate series
        industry_funds_df.columns = [f"{i}_{j}" for i, j in industry_funds_df.columns]
        # fix datetime index type issue
        # https://stackoverflow.com/questions/48248239/pandas-how-to-convert-rangeindex-into-datetimeindex
        industry_funds_df.index = pd.to_datetime(industry_funds_df.index)
        industry_funds_series = TimeSeries.from_dataframe(industry_funds_df, freq="B")
        industry_funds_series = industry_funds_series.slice(
            train_date_start, industry_funds_series.end_time()
        )
        filler = MissingValuesFiller(n_jobs=-1)
        industry_funds_filled = filler.transform(industry_funds_series)
        assert len(industry_funds_filled.gaps()) == 0
        industry_funds_series = industry_funds_filled
        # logger.debug(
        #     f"Finished preparing past covariates: industry funds. {len(industry_funds_series.columns)} columns, {len(industry_funds_series)} records,  \ncolumns: \n{industry_funds_series.columns}"
        # )
        return industry_funds_series

    def load_data(self, stock_tickers: set = None, start_date: pd.Timestamp = None):
        self.__start_date = start_date
        self.__load_tickers = stock_tickers
        self.load_past_covariates()
        self.load_future_covariates()
        self.data_loaded = True

    def load_past_covariates(self):
        self.load_earnings()
        self.load_key_metrics()
        self.load_broad_market()
        self.load_sectors()
        self.load_industry_funds()
        self.load_institutional_symbol_ownership()

    def load_institutional_symbol_ownership(self):
        inst_ownership_file = (
            "data/data-3rd-party/institutional_symbol_ownership.parquet"
        )
        logger.info(f"Loading data from: {inst_ownership_file}")
        # df = pd.read_csv(inst_ownership_file, low_memory=False)
        df = pd.read_parquet(inst_ownership_file, filters=self.pyarrow_filters)
        # logger.info("inst_symbol_ownership_df.columns", df.columns)
        # df["date"] = pd.to_datetime(df["date"])

        df = df.drop_duplicates()  # subset=["symbol", "date"])
        assert not df.duplicated().any()
        # df = df.set_index(keys=["symbol", "date"])
        assert df.index.has_duplicates == False
        self.inst_symbol_ownership_df = df

    def load_broad_market(self):
        file = "data/data-3rd-party/broad_market.parquet"
        # self.broad_market_df = pd.read_csv(csv_file, header=[0, 1], index_col=0)
        self.broad_market_df = pd.read_parquet(file)

    def load_sectors(self):
        file = "data/data-3rd-party/sectors.parquet"
        self.sectors_df = pd.read_parquet(file)

    def load_industry_funds(self):
        file = "data/data-3rd-party/industry_funds.parquet"
        self.industry_funds_df = pd.read_parquet(file)

    def load_future_covariates(self):
        self.load_estimates()
        self.load_dividends()
        self.load_splits()

    def prepare_data(
        self,
        stock_price_series: dict = None,
        target_columns: Union[str, list] = None,
        train_date_start: pd.Timestamp = None,
        min_samples: int = None,
        pred_horizon: int = None,
    ):
        logger.info("preparing model data")
        assert (
            self.data_loaded is True
        ), "Data needs to be loaded before it can be prepared. Make sure to first call load_data()"
        self.__train_date_start = train_date_start
        self.prepare_past_covariates(
            stock_price_series=stock_price_series,
            target_columns=target_columns,
            train_date_start=train_date_start,
        )
        self.prepare_future_covariates(
            stock_price_series=stock_price_series,
            min_samples=min_samples,
            pred_horizon=pred_horizon,
        )

    def load_earnings(self):
        earnings_file = "data/data-3rd-party/earnings_calendar.parquet"
        logger.info(f"Loading data from: {earnings_file}")
        # earnings_loaded_df = pd.read_csv(earnings_csv_file)
        logger.info(f"pyarrow_filters: {self.pyarrow_filters}")
        earnings_loaded_df = pd.read_parquet(
            earnings_file, filters=self.pyarrow_filters
        )
        # logger.info("earnings_loaded_df.columns", earnings_loaded_df.columns)
        # earnings_loaded_df["date"] = pd.to_datetime(earnings_loaded_df["date"])
        earnings_unique = (
            earnings_loaded_df.drop_duplicates()
        )  # subset=["Symbol", "Date"])
        assert not earnings_unique.duplicated().any()
        # earnings_unique = earnings_unique.set_index(keys=["symbol", "date"])
        earnings_unique = earnings_unique[
            ~earnings_unique.index.duplicated(keep="first")
        ]
        assert earnings_unique.index.has_duplicates == False
        logger.info(
            f"Loading a total of {len(earnings_unique)} unique earnings records"
        )
        self.earnings_loaded_df = earnings_unique

    def load_key_metrics(self):
        kms_file = "data/data-3rd-party/keymetrics_history.parquet"
        logger.info(f"Loading data from: {kms_file}")
        # kms_loaded_df = pd.read_csv(kms_file)
        kms_loaded_df = pd.read_parquet(kms_file, filters=self.pyarrow_filters)
        self.kms_loaded_df = kms_loaded_df

    def load_estimates(self):
        self.est_loaded_df = {}
        for period in fiscal_periods:
            assert period in fiscal_periods
            est_file = f"data/data-3rd-party/analyst_estimates_{period}.parquet"
            logger.info(f"Loading data from: {est_file}")
            # est_loaded_df = pd.read_csv(est_file)
            est_loaded_df = pd.read_parquet(est_file, filters=self.pyarrow_filters)
            # drop duplicate index rows
            est_loaded_df = est_loaded_df[~est_loaded_df.index.duplicated(keep="last")]
            assert est_loaded_df.index.is_unique
            est_loaded_df = est_loaded_df.drop_duplicates()
            assert not est_loaded_df.duplicated().any()
            # logger.info(f'{period} estimates prepared: \n{est_unique}')
            self.est_loaded_df[period] = est_loaded_df

    def load_dividends(self):
        self.dividends_loaded_df = {}
        data_file = f"{self.data_dir}/{self.data_3rd_party}/stock_dividends.parquet"
        logger.info(f"Loading data from: {data_file}")
        loaded_df = pd.read_parquet(data_file, filters=self.pyarrow_filters)
        assert loaded_df.index.is_unique
        loaded_df = loaded_df.drop_duplicates()
        assert not loaded_df.duplicated().any()
        # logger.debug(f"Dividends loaded: \n{loaded_df}")
        self.dividends_loaded_df = loaded_df

    def load_splits(self):
        self.splits_loaded_df = {}
        data_file = f"{self.data_dir}/{self.data_3rd_party}/stock_splits.parquet"
        logger.info(f"Loading data from: {data_file}")
        loaded_df = pd.read_parquet(data_file, filters=self.pyarrow_filters)
        assert loaded_df.index.is_unique
        loaded_df = loaded_df.drop_duplicates()
        assert not loaded_df.duplicated().any()
        # logger.debug(f"Splits loaded: \n{loaded_df}")
        self.splits_loaded_df = loaded_df

    def est_add_future_periods(self, est_df=None, n_future_periods=None, period=None):
        """
        Prepare time series with concatenated estimates for n_future_periods
        """
        # new_df = pd.DataFrame(index=est_df.index)
        # logger.info('est_df', est_df)
        # add fiscalDateEnding columns
        # fde = pd.to_datetime(est_df.index)
        fde_year = est_df.index.year
        fde_month = est_df.index.month
        fde_day = est_df.index.day
        n_cols = len(est_df.columns)
        est_expanded_df = est_df.copy()
        est_expanded_df.insert(
            loc=n_cols, column="fiscalDateEnding_year", value=fde_year
        )
        est_expanded_df.insert(
            loc=n_cols, column="fiscalDateEnding_month", value=fde_month
        )
        est_expanded_df.insert(loc=n_cols, column="fiscalDateEnding_day", value=fde_day)
        # shift date index by future period delta
        # so it can align and merge with the time series where the future period estimate columns will be used as past covariates
        est_shifted_df = est_expanded_df.copy()
        # https://pandas.pydata.org/docs/reference/api/pandas.Index.shift.html
        prange = range(1, n_future_periods + 1)
        assert est_shifted_df.index.is_unique == True
        est_shifted_df = est_shifted_df.shift(
            periods=prange, suffix=f"_p_{period}"
        )  # freq=fiscal_freq[period],
        # logger.info('est_shifted_df\n', est_shifted_df)
        # est_shifted_df.add_suffix(f'_p{n}')
        # new_df.join(est_shifted_df, how='outer', sort=True, validate='1:1')
        new_df = est_shifted_df
        # logger.info('new_df', new_df)
        return new_df

    def prepare_est_series(
        self,
        all_est_df=None,
        n_future_periods=None,
        period=None,
        stock_price_series=None,
    ):
        """
        Prepare future covariate series with analyst estimates for a given period (annual or quarter).
        :param all_est_df: estimates dataframe indexed by ['Symbol', 'Date']
        :param n_future_periods: number of periods of future estimates to make visible at each timeseries date
        :param period: quarter or annual
        :return: estimate series expanded with forward periods at each series date indexed row
        """
        logger.info(f"preparing future covariates: analyst estimates[{period}]")
        assert period in fiscal_periods
        t_est_series = {}
        for t, prices in stock_price_series.items():
            # logger.info(f'ticker {t}')
            try:
                est_df = all_est_df.loc[[t]].copy()
                est_df = est_df.droplevel("Symbol")
                est_df.index = pd.to_datetime(est_df.index)
                assert not est_df.index.duplicated().any()
                # expand series with estimates from future periods
                est_df = self.est_add_future_periods(
                    est_df=est_df, n_future_periods=n_future_periods, period=period
                )
                # logger.info(f'{t} estimates columns: \n{est_df.columns}')
                # align dates to business days
                est_df = self.df_index_to_biz_days(est_df)
                # there should be no duplicate rows
                est_df = est_df[~est_df.index.duplicated(keep="first")]
                assert not est_df.index.duplicated().any()
                # expand date index to match target price series dates and pad data
                est_series_tmp = TimeSeries.from_dataframe(
                    est_df, freq="B", fill_missing_dates=True
                )
                # logger.info(f'est_series_tmp start time, end time: {est_series_tmp.start_time()}, {est_series_tmp.end_time()}')
                est_df = est_series_tmp.pd_dataframe()
                # Make current annual/quarter period estimates available on all business days through end of the period
                est_df.ffill(inplace=True)
                est_ser = TimeSeries.from_dataframe(est_df, freq="B", fillna_value=-1)
                est_padded = self.pad_covs(
                    cov_series=est_ser, price_series=prices, fillna_value=-1
                )
                assert (
                    len(est_padded.gaps()) == 0
                ), f"found gaps in series: \n{est_padded.gaps()}"
                t_est_series[t] = est_padded

            except KeyError as e:
                logger.info(
                    f"""Skipping {t} from covariates series. 
                        No analyst estimates available for {t}, error: {type(e)}, {e}"""
                )
        return t_est_series

    def prepare_analyst_estimates(self, stock_price_series=None):
        logger.info("preparing future covariates: analyst estimates")
        quarter_est_series = self.prepare_est_series(
            all_est_df=self.est_loaded_df["quarter"],
            n_future_periods=4,
            period="quarter",
            stock_price_series=stock_price_series,
        )
        annual_est_series = self.prepare_est_series(
            all_est_df=self.est_loaded_df["annual"],
            n_future_periods=2,
            period="annual",
            stock_price_series=stock_price_series,
        )
        return quarter_est_series, annual_est_series

    def prepare_past_covariates(
        self,
        stock_price_series: dict = None,
        target_columns: Union[str, list] = None,
        train_date_start: pd.Timestamp = None,
    ):
        logger.info("preparing past covariates")
        # start with price-volume covariates
        past_covariates = self.get_price_covariates(
            stock_price_series=stock_price_series, target_columns=target_columns
        )
        # add revenue and earnings covariates
        earn_covariates = self.prepare_earn_series(tickers=stock_price_series.keys())
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=earn_covariates
        )
        # add key metrics covariates
        kms_series = self.prepare_key_metrics(stock_price_series=stock_price_series)
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=kms_series
        )
        # add historical stock splits covariates
        stock_split_series = self.prepare_splits(stock_price_series=stock_price_series)
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=stock_split_series
        )
        # add stock institutional ownership covariates
        inst_ownership_series = self.prepare_institutional_symbol_ownership_series(
            stock_price_series=stock_price_series
        )
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=inst_ownership_series
        )
        # add broad market covariates
        broad_market_series = self.prepare_broad_market_series(
            train_date_start=train_date_start
        )
        broad_market_dict = {t: broad_market_series for t in stock_price_series.keys()}
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=broad_market_dict
        )
        # add market sectors covariates
        sectors_series = self.prepare_sectors_series(train_date_start=train_date_start)
        sectors_dict = {t: sectors_series for t in stock_price_series.keys()}
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=sectors_dict
        )
        # add industry fund covariates
        industry_funds_series = self.prepare_industry_fund_series(
            train_date_start=train_date_start
        )
        industry_funds_dict = {
            t: industry_funds_series for t in stock_price_series.keys()
        }
        past_covariates = self.stack_covariates(
            old_covs=past_covariates, new_covs=industry_funds_dict
        )
        self.past_covariates = past_covariates
        logger.info(f"Prepared past covariates.")

    def __add_holidays(self, series_dict: dict = None):
        logger.info("preparing future covariates: holidays")
        new_series = {}
        for t, series in series_dict.items():
            series_with_holidays = series.add_holidays(country_code="US")
            new_series[t] = series_with_holidays
            # logger.info(f'ticker: {t} , {ticker_series[t]}')
        return new_series

    def __extend_series(self, n: int = -1, series: {} = None, target: {} = None):
        new_series = {}
        for t, s in series.items():
            # logger.info(
            #     f"{t} series before extension start, end: {s.start_time()}, {s.end_time()}"
            # )
            # logger.info(f"target {t} end time: {target[t].end_time()}")
            start = s.start_time()
            if s.end_time() > target[t].end_time() + BDay(n=n):
                new_series[t] = s
                # logger.info(
                #     f"No need to extend {t} series. End greater than target end."
                # )
            else:
                end = s.end_time() + BDay(n=n)
                df = s.pd_dataframe()
                idx = pd.date_range(start=start, end=end, freq="B")
                df = df.reindex(idx).ffill()
                s_ext = TimeSeries.from_dataframe(df, freq="B", fill_missing_dates=True)
                new_series[t] = s_ext
                # logger.info(
                #     f"{t} series after extension start, end: {s_ext.start_time()}, {s_ext.end_time()}"
                # )
        return new_series

    def prepare_dividends(self, stock_price_series: dict = None):

        def no_dividends_stock(start_date: pd.Timestamp):
            """Prepare placeholder dataframe for a stock without any historical dividends"""
            return pd.DataFrame(
                {
                    "dividend": [0],
                    "recordDate": [start_date],
                    "paymentDate": [start_date],
                    "declarationDate": [start_date],
                }
            )

        logger.info("preparing future covariates: stock dividends")
        # convert date strings to numerical representation
        div_df = self.dividends_loaded_df.copy()
        # logger.debug(f"div_df sample: \n{div_df}")
        # logger.debug(f"div_df.columns: {div_df.columns}")
        # remove adjDividend because it cannot be used as a future covariate
        # remove label because it is simply another string format for Date,
        # which does not add unique and useful information to the training data
        div_df.drop(["adjDividend", "label"], axis="columns", inplace=True)
        # convert date columns from string to pd.Timestamp
        for c in ["declarationDate", "paymentDate", "recordDate"]:
            div_df[c] = pd.to_datetime(div_df[c])
        # drop rows with null / unknown declarationDate, paymentDate or dividend
        div_df.dropna(
            subset=["declarationDate", "paymentDate", "dividend"],
            how="any",
            inplace=True,
        )
        t_series = {}
        for t, prices in stock_price_series.items():
            try:
                # logger.debug(f"ticker: {t}")
                try:
                    t_div = div_df.loc[[t]]
                except KeyError as e:
                    # no dividends information available for ticker
                    # assume stock has never issued dividends
                    t_div = no_dividends_stock(start_date=prices.start_time())
                # logger.debug(f"t_div sample for {t}: \n{t_div}")
                # replace ["Symbol", "Date"] index with declarationDate
                # simultaneously drop previous ["Symbol", "Date"] multiindex columns
                t_div = t_div.set_index("declarationDate", drop=True)
                # logger.debug(f"t_div.columns: {t_div.columns}")
                # logger.debug(f"t_div sample for {t}: \n{t_div}")
                t_div.index.name = "Date"
                # logger.debug(f"t_div sample for {t}: \n{t_div}")
                # logger.debug(
                # f"t_div  {t} samples with null index: \n{t_div[t_div.index.isnull()]}"
                # )
                assert not t_div.index.isnull().any()
                date_col_names = ["paymentDate", "recordDate"]
                for date_col_name in date_col_names:
                    date_col = t_div[date_col_name]
                    # date_col = pd.to_datetime(t_div[date_col_name])
                    date_col_year = date_col.dt.year
                    date_col_month = date_col.dt.month
                    date_col_day = date_col.dt.day
                    div_n_cols = len(t_div.columns)
                    # add encodings for year, month and day to make it easier for the model to see how far in the future is the next dividends date
                    # and potentially discover cyclical behavior in stock price movement based on dividends
                    t_div.insert(
                        loc=div_n_cols,
                        column=f"{date_col_name}_day",
                        value=date_col_day,
                    )
                    t_div.insert(
                        loc=div_n_cols,
                        column=f"{date_col_name}_month",
                        value=date_col_month,
                    )
                    t_div.insert(
                        loc=div_n_cols,
                        column=f"{date_col_name}_year",
                        value=date_col_year,
                    )
                t_div.drop(date_col_names, axis="columns", inplace=True)
                # logger.debug(f"t_div sample for {t}: \n{t_div}")
                # logger.debug(f"t_div.index for {t}: \n{t_div.index}")
                # align declaration dates to the closest business days used in the target timeseries index
                t_div = self.df_index_to_biz_days(t_div)
                # drop rows with duplicate datetime index values
                t_div = t_div[~t_div.index.duplicated(keep="last")]
                assert (
                    t_div.index.is_unique
                ), f"Duplicate index keys found for {t}: {t_div[t_div.index.duplicated()]}"
                # make sure we don't leak forward dividend declarations
                # that are not known at prediction time
                t_div.drop(
                    index=t_div[t_div.index > prices.end_time()].index, inplace=True
                )
                assert not t_div.empty
                assert (
                    t_div.index.max() <= prices.end_time()
                ), f"""Dividend declarations which are not know at prediction time must not leak into future covariates: 
                    {t} dividends end time {t_div.index.max()} > price target end time {prices.end_time()}"""
                # logger.debug(f"t_div sample for {t}: \n{t_div}")
                tmp = TimeSeries.from_dataframe(
                    t_div, freq="B", fill_missing_dates=True
                )
                # logger.debug(
                #     f"tmp series start time, end time: {tmp.start_time()}, {tmp.end_time()}"
                # )
                t_div = tmp.pd_dataframe()
                t_div.ffill(inplace=True)
                # Fill empty cells with 0 to indicate to the model unknown dividend
                ts = TimeSeries.from_dataframe(t_div, freq="B", fillna_value=0)
                ts_padded = self.pad_covs(
                    cov_series=ts, price_series=prices, fillna_value=0
                )
                assert (
                    len(ts_padded.gaps()) == 0
                ), f"Found unexpected gaps in series: \n{ts_padded.gaps()}"
                # logger.debug(f"ts_padded sample for {t}: \n{ts_padded.pd_dataframe()}")
                # logger.debug(
                #     f"ts_padded series start time, end time: {ts_padded.start_time()}, {ts_padded.end_time()}"
                # )
                t_series[t] = ts_padded
            except (KeyError, ValueError, AssertionError) as e:
                logger.exception(f"Skipping {t} due to error: {type(e)}: {e}")

        # if len(t_series.keys()) > 0:
        #     logger.debug(f"t_series.columns[{t}]: {t_series[t].columns}")
        #     logger.debug(f"len(t_series[{t}]): {len(t_series[t])}")
        #     logger.debug(f"t_series[{t}]: {t_series[t]}")
        #     logger.debug(
        #         f"t_series[{t}] start time, end time: {t_series[t].start_time()}, {t_series[t].end_time()}"
        #     )
        return t_series

    def prepare_splits(self, stock_price_series: dict = None):

        def no_splits_stock(start_date: pd.Timestamp):
            """Prepare placeholder dataframe for a stock without any historical splits"""
            df = pd.DataFrame(
                {
                    "Symbol": ["any"],
                    "Date": [start_date],
                    "numerator": [1],
                    "denominator": [1],
                }
            )
            df = df.set_index(keys=["Symbol", "Date"], drop=True)
            # logger.debug(f"df: \n{df}")
            return df

        logger.info("preparing past covariates: stock splits")
        # convert date strings to numerical representation
        splits_df = self.splits_loaded_df.copy()
        # logger.debug(f"self.splits_loaded_df.columns: {self.splits_loaded_df.columns}")
        # remove label because it is simply another string format for Date,
        # which does not add unique and useful information to the training data
        splits_df = splits_df.drop(["label"], axis="columns")
        # logger.debug(f"splits_df.columns {splits_df.columns}")
        # logger.debug(f"splits_df sample: \n{splits_df}")
        t_series = {}
        for t, prices in stock_price_series.items():
            try:
                # logger.debug(f"ticker: {t}")
                try:
                    t_splits = splits_df.loc[[t]]
                except KeyError as e:
                    # no dividends information available for ticker
                    # assume stock has never issued dividends
                    t_splits = no_splits_stock(start_date=prices.start_time())

                # logger.debug(f"t_splits sample: \n{t_splits}")
                t_splits = t_splits.droplevel("Symbol")
                # t_splits.index = pd.to_datetime(t_splits.index)
                # logger.debug(f"index type for {t}: {type(t_splits.index)}")
                assert t_splits.index.is_unique
                assert not t_splits.index.isnull().any()
                # align declaration dates to the closest business days used in the target timeseries index
                t_div = self.df_index_to_biz_days(t_splits)
                # drop rows with duplicate datetime index values
                t_splits = t_splits[~t_splits.index.duplicated(keep="last")]
                # logger.debug(f"t_splits sample for {t}: {t_splits}")
                tmp = TimeSeries.from_dataframe(
                    t_splits, freq="B", fill_missing_dates=True
                )
                # logger.debug(
                #     f"tmp series start time, end time: {tmp.start_time()}, {tmp.end_time()}"
                # )
                t_splits = tmp.pd_dataframe()
                # Do not interpolate splits forward, because
                # the Date index is the date when the split occurs
                # the model should not be confused as to when exactly
                # a stock split takes place
                # --> NO: t_splits.ffill(inplace=True)
                # fill in empty numerator and denominator values with 1s,
                # which is equivalent to no stock splits occuring on a given date
                ts = TimeSeries.from_dataframe(t_div, freq="B", fillna_value=1)
                ts_padded = self.pad_covs(
                    cov_series=ts, price_series=prices, fillna_value=1
                )
                assert (
                    len(ts_padded.gaps()) == 0
                ), f"found gaps in series: \n{ts_padded.gaps()}"
                # logger.debug(f"ts_padded sample for {t}: {ts_padded.pd_dataframe()}")
                t_series[t] = ts_padded
            except (KeyError, ValueError, AssertionError) as e:
                logger.exception(f"Skipping {t} due to error: {type(e)}: {e}")

        return t_series

    def prepare_future_covariates(
        self,
        stock_price_series: dict = None,
        min_samples=None,
        pred_horizon: int = None,
    ):
        logger.info("Preparing future covariates")
        # add analyst estimates
        quarter_est_series, annual_est_series = self.prepare_analyst_estimates(
            stock_price_series=stock_price_series
        )
        ## Stack analyst estimates to future covariates
        future_covariates = self.stack_covariates(
            old_covs=quarter_est_series,
            new_covs=annual_est_series,
            min_samples=min_samples,
        )
        # add stock dividends covariates
        # for some stocks there will be dividends declared prior to the forecast date
        # with known payment dates scheduled into the future
        dividend_series = self.prepare_dividends(stock_price_series=stock_price_series)
        future_covariates = self.stack_covariates(
            old_covs=future_covariates,
            new_covs=dividend_series,
            min_samples=min_samples,
        )
        # extend covars into forecast horizon future dates
        future_covariates = self.__extend_series(
            n=pred_horizon, series=future_covariates, target=stock_price_series
        )
        # add holidays
        future_covariates = self.__add_holidays(future_covariates)
        # update covs object state
        self.future_covariates = future_covariates



from typing import Union
from loguru import logger
from darts.dataprocessing.transformers import MissingValuesFiller
from darts import TimeSeries
import pandas as pd


class Targets:
    def __init__(self) -> None:
        self.target_series = {}

    def load_data(
        self,
        stock_tickers: set = None,
        min_samples: int = -1,
        start_date: pd.Timestamp = None,
    ):
        self.__start_date = start_date
        self.__load_tickers = stock_tickers
        self.min_samples = min_samples
        self.load_stock_prices()

    @property
    def pyarrow_filters(self):
        return [
            ("Symbol", "in", self.__load_tickers),
            ("Date", ">=", self.__start_date),
        ]

    def load_stock_prices(self):
        stocks_price_file = "data/data-3rd-party/all_stocks_price_hist_1d.parquet"
        logger.info(
            f"Loading data from: {stocks_price_file} with filter: {self.pyarrow_filters}"
        )
        # load into a dataframe with valid market calendar days
        # stocks_df = pd.read_csv(
        #     stocks_price_file, header=[0, 1], index_col=0, on_bad_lines="warn"
        # )
        stocks_df = pd.read_parquet(
            stocks_price_file,
            filters=self.pyarrow_filters,
            dtype_backend="numpy_nullable",
        )
        logger.info("filtered data loaded")
        stocks_df = stocks_df.dropna()
        # stocks_df.index = pd.to_datetime(stocks_df.index)
        stock_price_dict = {}
        # stock_tickers = self.__get_stock_tickers(stocks_df)
        tickers = list(stocks_df.index.levels[0])
        logger.info(f"price history loaded for {len(tickers)} stocks: \n{tickers}")
        for t in tickers:
            # logger.info(f"validating price data for {t}")
            stock_full_hist = stocks_df.loc[[t]]
            if len(stock_full_hist.index) >= self.min_samples:
                stock_full_hist = stock_full_hist.droplevel("Symbol")
                stock_full_hist.index = pd.to_datetime(stock_full_hist.index)
                # Drop Adj Close because yfiance changes its values retroactively at future dates
                # after stock dividend or split dates, which makes training data less stable
                # Ref: https://help.yahoo.com/kb/adjusted-close-sln28256.html
                stock_price_dict[t] = stock_full_hist.drop(columns=["Adj Close"])
                # logger.info(f'ticker: {t}')
                # logger.info(f'ticker historic data: {ticker_dict[t]}')
            else:
                logger.info(
                    f"Skipping {t} from price series. Not enough samples for model training."
                )
        self.stock_price_dict = stock_price_dict

    def prepare_data(
        self, stock_price_series: dict = None, target_columns: Union[str, list] = None
    ):
        def drop_non_target_columns(series):
            cols = series.columns
            non_target_columns = list(set(cols) - set(target_columns))
            new_series = series.drop_columns(col_names=non_target_columns)
            # logger.info(f'dropped non-target columns: {non_target_columns}')
            return new_series

        if isinstance(target_columns, list) and len(target_columns) == 1:
            target_columns = target_columns[0]
            logger.info(f"Single target column selected: {target_columns}")

        if isinstance(target_columns, str):
            # prepare target univariate series for Close price
            target_series = {
                t: stock_price_series[t].univariate_component(target_columns)
                for t in stock_price_series.keys()
            }
            logger.info(f"Preparing univariate target series: {target_columns}")
        else:
            # prepare target multivariate series for Open, Close and Volume
            target_series = {
                t: drop_non_target_columns(s) for t, s in stock_price_series.items()
            }
            logger.info(f"Preparing multivariate target series: {target_columns}")
        self.target_series = target_series

    def prepare_stock_price_series(self, train_date_start: pd.Timestamp = None):
        loaded_tickers = self.stock_price_dict.keys()
        logger.info(
            f"Preparing ticker series for {len(loaded_tickers)} stocks: \n{loaded_tickers}"
        )
        stock_price_series = {
            t: TimeSeries.from_dataframe(self.stock_price_dict[t], freq="B")
            for t in loaded_tickers
        }
        logger.info("Ticker series dict created.")
        filler = MissingValuesFiller()
        for t, series in stock_price_series.items():
            # gaps = series.gaps(mode="any")
            # logger.info(f'ticker: {t} gaps: \n {gaps}')
            series_filled = filler.transform(series)
            # check for any data gaps
            price_gaps = series_filled.gaps(mode="any")
            assert len(price_gaps) == 0
            # logger.info(f'ticker: {t} gaps after filler: \n {any_price_gaps}')
            stock_price_series[t] = series_filled
        logger.info("Filled missing values in ticker series.")
        for t, series in stock_price_series.items():
            stock_price_series[t] = series.slice(train_date_start, series.end_time())
            # logger.info(f'ticker: {t} , {ticker_series[t]}')
        # add holidays as future covariates
        logger.info("Aligned ticker series dict with train start date.")
        logger.info("Ticker series prepared.")
        return stock_price_series



"""
Class representing an advanced tab for querying and scanning forecasts.

Attributes:
    canswim_model (CanswimModel): An instance of the CanswimModel class.

Methods:
    __init__(canswim_model): Initializes the AdvancedTab class.
    scan_forecasts(query): Scans the forecasts based on the provided query.

"""

from loguru import logger
from canswim.model import CanswimModel
import gradio as gr
import duckdb


class AdvancedTab:

    def __init__(self, canswim_model: CanswimModel = None, db_path=None):
        assert canswim_model is not None
        assert db_path is not None
        self.canswim_model = canswim_model
        self.db_path = db_path
        with gr.Row():
            self.queryBox = gr.TextArea(
                value="""SELECT 
                    f.symbol, 
                    f.start_date as forecast_start_date, 
                    arg_max(c.close, c.date) as prior_close_price, 
                    min("close_quantile_0.2") as forecast_close_low, 
                    max("close_quantile_0.5") as forecast_close_high,
                    100*(forecast_close_high / prior_close_price - 1) as reward_percent,
                    (forecast_close_high - prior_close_price)/GREATEST(prior_close_price-forecast_close_low, 0.01) as reward_risk,
                    max(e.mal_error) as backtest_error,
                    max(c.date) as prior_close_date,
                FROM forecast f, close_price c, backtest_error as e, latest_forecast as lf
                WHERE f.symbol = lf.symbol AND
                    f.symbol = c.symbol AND f.symbol = e.symbol AND c.date < lf.date
                GROUP BY f.symbol, f.start_date, c.symbol, e.symbol, lf.symbol, lf.date
                HAVING forecast_close_high > prior_close_price AND
                    f.start_date = lf.date AND
                    reward_risk> 3 AND reward_percent >= 20
                    and forecast_start_date = max(lf.date)
            """
            )

        with gr.Row():
            self.runBtn = gr.Button(value="Run Query", variant="primary")

        with gr.Row():
            self.queryResult = gr.Dataframe()

        self.runBtn.click(
            fn=self.scan_forecasts,
            inputs=[
                self.queryBox,
            ],
            outputs=[self.queryResult],
        )
        self.queryBox.submit(
            fn=self.scan_forecasts,
            inputs=[
                self.queryBox,
            ],
            outputs=[self.queryResult],
        )

    def scan_forecasts(self, query):
        try:
            # only run select queries
            if query.strip().upper().startswith("SELECT"):
                with duckdb.connect(self.db_path) as db_con:
                    sql_result = db_con.sql(query)
                    logger.info(f"SQL Result: \n{sql_result}")
                    df = sql_result.df()
                    df["prior_close_date"] = df["prior_close_date"].dt.strftime(
                        "%Y-%m-%d"
                    )
                    df["forecast_start_date"] = df["forecast_start_date"].dt.strftime(
                        "%Y-%m-%d"
                    )
                    return df
        except Exception as e:
            gr.Error("An error occurred while running the query:", e)
            return None



#!/usr/bin/env python
"""
   Copyright 2024 Cocoonhive LLC

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import signal
import sys
from dotenv import load_dotenv

# load os env vars before loguru import
# otherwise it won't pickup LOGURU_LEVEL
load_dotenv(override=True)

from loguru import logger
import os
import argparse
from canswim import dashboard, gather_data, model_search, train, forecast
from canswim.hfhub import HFHub

# Instantiate the parser
parser = argparse.ArgumentParser(
    prog="canswim",
    description="""CANSWIM is a toolkit for CANSLIM style investors.
        Aims to complement the Simple Moving Average and other technical indicators.
        """,
    epilog="NOTE: NOT FINANCIAL OR INVESTMENT ADVICE. USE AT YOUR OWN RISK.",
)

parser.add_argument(
    "task",
    type=str,
    help="""Which %(prog)s task to run:
        `dashboard` for stock charting and scans of recorded forecasts.
        'gatherdata` to gather 3rd party stock market data and save to HF Hub.
        'downloaddata` download model training and forecast data from HF Hub to local data storage.
        'uploaddata` upload to HF Hub any interim changes to local train and forecast data.
        `modelsearch` to find and save optimal hyperparameters for model training.
        `train` for continuous model training.
        `forecast` to run forecast on stocks and upload dataset to HF Hub.
        """,
    choices=[
        "dashboard",
        "gatherdata",
        "downloaddata",
        "uploaddata",
        "modelsearch",
        "train",
        "forecast",
    ],
)

parser.add_argument(
    "--forecast_start_date",
    type=str,
    required=False,
    help="""Optional argument for the `forecast` task. Indicate forecast start date in YYYY-MM-DD format. If not specified, forecast will start from the end of the target series.""",
)

parser.add_argument(
    "--new_model",
    type=bool,
    required=False,
    default=False,
    help="""Optional argument for the `train` task. Whether to train a newly created model or continue training an existing pre-trained model.""",
)

parser.add_argument(
    "--same_data",
    type=bool,
    required=False,
    default=False,
    help="""Optional argument for the `dashboard` task. Whether to reuse previously created search database (faster start time) or update with new forecast data (slower start time).""",
)

args = parser.parse_args()

logging_dir = os.getenv("logging_dir", "tmp")
logging_path = f"{logging_dir}/canswim.log"
rot = "24 hours"
ret = "30 days"
logger.add(logging_path, rotation=rot, retention=ret)

logger.info(
    "Logging to: {p} with rotation {rot} and retention {ret}",
    p=logging_path,
    rot=rot,
    ret=ret,
)

logger.info("command line args: {args}", args=args)

hfhub = HFHub()
load_dotenv(override=True)
repo_id = os.getenv("repo_id", "ivelin/canswim")


def signal_handler(sig, frame):
    print("Ctrl+C - Exit")
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)

match args.task:
    case "dashboard":
        dashboard.main(same_data=args.same_data)
    case "modelsearch":
        model_search.main()
    case "gatherdata":
        gather_data.main()
    case "downloaddata":
        hfhub.download_data()
    case "uploaddata":
        hfhub.upload_data()
    case "train":
        train.main(new_model=args.new_model)
    case "forecast":
        forecast.main(forecast_start_date=args.forecast_start_date)
    case _:
        logger.error("Unrecognized task argument {m} ", m=args.module)



from canswim.model import CanswimModel
from darts.models import TiDEModel
from canswim.hfhub import HFHub
from darts.metrics import quantile_loss
import numpy as np
import os
from dotenv import load_dotenv
import pandas as pd
from pandas.tseries.offsets import Day, Week
from loguru import logger


class CanswimTrainer:

    def __init__(self) -> None:
        self.canswim_model = CanswimModel()

    def prepare_data(self):
        self.canswim_model.prepare_data()
        self.canswim_model.plot_splits()
        self.canswim_model.plot_seasonality()

    def build_new_model(self):
        """Build a new model using known optimal hyperparameters"""
        self.canswim_model.build(
            input_chunk_length=252,
            output_chunk_length=42,
            hidden_size=2048,
            num_encoder_layers=3,
            num_decoder_layers=2,
            decoder_output_dim=8,
            temporal_decoder_hidden=80,
            use_layer_norm=True,
            use_reversible_instance_norm=True,
            dropout=0.3,
            optimizer_kwargs={"lr": 1e-05},
            save_checkpoints=False,  # checkpoint to retrieve the best performing model state,
            force_reset=False,
        )

    def plot_backtest_results(self):
        # backtest on 3 stocks
        for i in range(min(len(self.canswim_model.targets_list), 3)):
            start_list = self.canswim_model.get_val_start_list()
            backtest = self.canswim_model.backtest(
                target=self.canswim_model.targets_list[i],
                start=start_list[i],
                past_covariates=self.canswim_model.past_cov_list[i],
                future_covariates=self.canswim_model.future_cov_list[i],
                forecast_horizon=self.canswim_model.pred_horizon,
            )
            # logger.info(f"target series: \n{target}")
            # logger.info(f"backtest series: \n{backtest}")
            loss_vals = []
            for p, b in enumerate(backtest):
                loss = quantile_loss(
                    self.canswim_model.targets_list[i], b, n_jobs=-1, verbose=True
                )
                logger.info(f"quantile loss: {loss} at prediction step {p}")
                loss_vals.append(loss)
            mean_loss = np.mean(loss_vals)
            logger.info(
                f"Mean Backtest Quantile Loss across all prediction periods: {mean_loss}"
            )
            self.canswim_model.plot_backtest_results(
                target=self.canswim_model.targets_list[i],
                backtest=backtest,
                start=start_list[i],
                forecast_horizon=self.canswim_model.pred_horizon,
            )


# main function
def main(new_model: bool = False):

    trainer = CanswimTrainer()

    n_outer_train_loop = 1
    hfhub = HFHub()
    default_repo_id = "ivelin/canswim"
    repo_id = default_repo_id

    def get_env():
        load_dotenv(override=True)
        nonlocal n_outer_train_loop
        n_outer_train_loop = int(os.getenv("n_outer_train_loop", 1))
        nonlocal repo_id
        repo_id = os.getenv("repo_id", default_repo_id)

    get_env()

    logger.info("n_outer_train_loop:", n_outer_train_loop, type(n_outer_train_loop))

    # build a new model or download existing model from hf hub or local dir
    if new_model:
        trainer.build_new_model()
    else:
        trainer.canswim_model.download_model(
            repo_id=repo_id
        )  # prepare next sample subset

    # download market data from hf hub if it hasn't been downloaded already
    trainer.canswim_model.download_data(repo_id=repo_id)

    wall_time = pd.Timestamp.now
    # set to yesterday
    yesterday = wall_time().floor(freq="D") - Day(n=1)
    # set to previous Saturday
    previous_weekend = wall_time().floor(freq="D") - Day(n=wall_time().day_of_week + 2)

    # train loop
    i = 0
    while i < n_outer_train_loop:
        logger.info(f"Outer train loop: {i}")
        try:
            # load a new data sample from local storage
            trainer.canswim_model.load_data()
            # prepare timeseries for training
            trainer.canswim_model.prepare_data()
            # train model
            trainer.canswim_model.train()
        except Exception as e:
            logger.exception(f"Skipping train loop due to ERROR.: {e}")
        # Daily routine
        if wall_time() > yesterday + Day(n=1):
            try:
                logger.info(
                    f"Starting daily routine. Yesterday: {yesterday}, wall_time: {wall_time()}"
                )
                # update yesterday marker to today
                yesterday = wall_time().floor(freq="D")
                # push to hf hub
                trainer.canswim_model.upload_model(repo_id=repo_id)
                # load back latest model weights from hf hub
                trainer.canswim_model.download_model(
                    repo_id=repo_id
                )  # prepare next sample subset
                logger.info("Finished daily routine.")
            except Exception as e:
                logger.exception("ERROR during daily routine: ", e)
        else:
            logger.info(
                f"Skipping daily routine. Yesterday: {yesterday}, wall_time: {wall_time()}"
            )
        # Weekend routine
        if wall_time() > previous_weekend + Week(n=1):
            try:
                logger.info(
                    f"Starting weekend routine. Previous weekend: {previous_weekend}, wall_time: {wall_time()}"
                )
                # update weekend marker to this week's Satuday
                previous_weekend = wall_time().to_period("W").start_time + Day(n=5)
                # gather_market_data() - in range from last gather until now
                # upload_data() - upload changed parquet files
                # run_forecasts() - run model forecast on all stocks which are far enough from previous and next earnings report, save parquet files
                # upload_forecasts() - upload changed parquet files
                logger.info("Finished weekend routine.")
            except Exception as e:
                logger.exception("ERROR during weekend routine: ", e)
        else:
            logger.info(
                f"Skipping weekend routine. Previous weekend: {previous_weekend}, wall_time: {wall_time()}"
            )
        # refresh any config changes in the OS environment
        i += 1
        get_env()
        # Note: load locally saved model
        # since PyTorch Lightning currently does not recommend
        # using fit() multiple times without loading model weights
        # WARNING:darts.models.forecasting.torch_forecasting_model:Attempting to retrain/fine-tune the model without resuming from a checkpoint. This is currently discouraged. Consider model `TiDEModel.load_weights()` to load the weights for fine-tuning.
        trainer.canswim_model.load_model()


if __name__ == "__main__":
    main()



"""CANSWIM Playground. A gradio app intended to be deployed on HF Hub."""

from canswim.model import CanswimModel
import pandas as pd
import gradio as gr
from canswim.hfhub import HFHub
from loguru import logger
from canswim.dashboard.charts import ChartTab
from canswim.dashboard.scans import ScanTab
from canswim.dashboard.advanced import AdvancedTab
from pandas.tseries.offsets import BDay
import duckdb
import os

# Note: It appears that gradio Plot ignores the backend plot lib setting
# pd.options.plotting.backend = "plotly"
# pd.options.plotting.backend = "matplotlib"
# pd.options.plotting.backend = "hvplot"

repo_id = "ivelin/canswim"


class CanswimPlayground:

    def __init__(self, same_data=False):
        self.same_data = same_data
        self.canswim_model = CanswimModel()
        self.hfhub = HFHub()
        data_dir = os.getenv("data_dir", "data")
        db_file = os.getenv("db_file", "local.duckdb")
        self.db_path = f"{data_dir}/{db_file}"
        forecast_subdir = os.getenv("forecast_subdir", "forecast/")
        self.forecast_path = f"{data_dir}/{forecast_subdir}"
        data_3rd_party = os.getenv("data-3rd-party", "data-3rd-party")
        price_data = os.getenv("price_data", "all_stocks_price_hist_1d.parquet")
        self.stocks_price_path = f"{data_dir}/{data_3rd_party}/{price_data}"
        stock_tickers_list = os.getenv("stock_tickers_list", "all_stocks.csv")
        self.stock_tickers_path = f"{data_dir}/{data_3rd_party}/{stock_tickers_list}"

    def download_model(self):
        """Load model from HF Hub"""
        # download model from hf hub
        logger.info(f"Downloading model from remote repo: {repo_id}")
        self.canswim_model.download_model(repo_id=repo_id)
        if self.canswim_model.torch_model is None:
            self.canswim_model.build()
        logger.info(f"trainer params {self.canswim_model.torch_model.trainer_params}")
        self.canswim_model.torch_model.trainer_params["logger"] = False

    def download_data(self):
        """Prepare time series for model forecast"""
        # download raw data from hf hub
        logger.info(f"Downloading data from remote repo: {repo_id}")
        self.hfhub.download_data(repo_id=repo_id)

    def initdb(self):
        with duckdb.connect(self.db_path) as db_con:
            create_new_db = True
            if self.same_data:
                result = db_con.table("stock_tickers").fetchone()
                if result is not None:
                    create_new_db = False
            if not create_new_db:
                logger.info("Reusing search database")
            else:
                logger.info("Creating search optimized database")
                db_con.sql(
                    f"""--sql
                    SET enable_progress_bar = true;        
                    """
                )
                logger.info("Creating stock_tickers table")
                db_con.sql(
                    f"""--sql
                    CREATE OR REPLACE TABLE stock_tickers 
                    AS SELECT * FROM read_csv('{self.stock_tickers_path}', header=True)
                    """
                )
                db_con.table("stock_tickers").show()
                db_con.sql(
                    f"""--sql
                    CREATE UNIQUE INDEX stock_tickers_sym_idx ON stock_tickers (symbol)
                    """
                )
                logger.info("Creating forecast table")
                db_con.sql(
                    f"""--sql
                    CREATE OR REPLACE TABLE forecast 
                    AS SELECT date, symbol, make_date(forecast_start_year, forecast_start_month, forecast_start_day) as start_date, COLUMNS(\"close_quantile_\d+\.\d+\") 
                    FROM read_parquet('{self.forecast_path}/**/*.parquet', hive_partitioning = 1) as f
                    SEMI JOIN stock_tickers
                    ON f.symbol = stock_tickers.symbol
                    """
                )
                db_con.table("forecast").show()
                db_con.sql(
                    f"""--sql
                    CREATE UNIQUE INDEX forecast_symd_idx
                    ON forecast (symbol, start_date, date)
                    """
                )
                logger.info("Creating latest_forecast table")
                db_con.sql(
                    f"""--sql
                    CREATE OR REPLACE TABLE latest_forecast AS
                        SELECT symbol, max(start_date) as date
                        FROM forecast as f
                        SEMI JOIN stock_tickers
                        ON f.symbol = stock_tickers.symbol                
                        GROUP BY symbol
                    """
                )
                db_con.table("latest_forecast").show()
                db_con.sql(
                    f"""--sql
                    CREATE UNIQUE INDEX latest_forecast_symd_idx
                    ON latest_forecast (symbol, date)
                    """
                )
                logger.info("Creating close_price table")
                db_con.sql(
                    f"""--sql
                    CREATE OR REPLACE TABLE close_price
                    AS SELECT Date, Symbol, "{self.canswim_model.target_column}" as Close
                    FROM read_parquet('{self.stocks_price_path}') as cp
                    SEMI JOIN stock_tickers
                    ON cp.symbol = stock_tickers.symbol;
                    """
                )
                db_con.table("close_price").show()
                db_con.sql(
                    f"""--sql
                    CREATE UNIQUE INDEX close_price_symd_idx
                    ON close_price (symbol, date)
                    """
                )
                logger.info("Creating backtest_error table")
                db_con.sql(
                    f"""--sql
                    CREATE OR REPLACE TABLE backtest_error 
                    AS SELECT f.symbol, mean(abs(log(greatest(f."close_quantile_0.5", 0.01)/cp.Close))) as mal_error
                    FROM forecast as f, close_price as cp
                    WHERE cp.symbol = f.symbol AND cp.date = f.date
                    GROUP BY f.symbol, cp.symbol
                    HAVING cp.symbol = f.symbol
                    """
                )
                db_con.table("backtest_error").show()
                db_con.sql(
                    f"""--sql
                    CREATE UNIQUE INDEX backtest_error_sym_idx
                    ON backtest_error (symbol)
                    """
                )

    def launch(self):
        self.download_model()
        self.download_data()
        self.initdb()
        with gr.Blocks(theme=gr.themes.Soft()) as demo:
            gr.Markdown(
                """
            CANSWIM Playground for CANSLIM style investors.
            * __NOT FINANCIAL OR INVESTMENT ADVICE. USE AT YOUR OWN RISK!__
            * Model trainer source repo [here](https://github.com/ivelin/canswim). Feedback welcome via github issues.
            """
            )
            with gr.Tab("Charts"):
                charts_tab = ChartTab(
                    canswim_model=self.canswim_model,
                    db_path=self.db_path,
                )
            with gr.Tab("Scans"):
                ScanTab(
                    self.canswim_model,
                    db_path=self.db_path,
                )
            with gr.Tab("Advanced Queries"):
                AdvancedTab(
                    self.canswim_model,
                    db_path=self.db_path,
                )

            demo.load(
                fn=charts_tab.plot_forecast,
                inputs=[charts_tab.tickerDropdown, charts_tab.lowq],
                outputs=[charts_tab.plotComponent, charts_tab.rrTable],
            )

        demo.queue().launch()


def main(same_data=False):

    canswim_playground = CanswimPlayground(same_data=same_data)
    canswim_playground.launch()


if __name__ == "__main__":
    main()
